package cz.cvut.fit.xml.index.tpa;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.jdom2.Document;
import org.jdom2.Element;

/**
 *
 * Construct <code>TreePathsAutomaton</code> for given XML document.
 *
 * @author sestaeli
 */
public class AutomatonFactory implements Serializable {

    //contains node information about its label, arity and sjt
    private ArrayList<Node> nodes;
    //mapping element label to its occurrences in an XML data tree
    private HashMap<String, TreeSet<Integer>> occurrences;
    //nondeterministic transition function
    private HashMap<Integer, HashMap<EdgeLabel, TreeSet<Integer>>> nDelta;
    //resulting automaton to be constructed
    private TreePathsAutomaton automaton;

    /**
     * Build <code>TreePathsAutomaton</code> for given XML document.
     *
     * @param document XML document to be indexed
     * @return resultant TreePathsAutomaton object
     */
    public TreePathsAutomaton build(Document document) {
        automaton = new TreePathsAutomaton();
        nDelta = new HashMap<>();
        occurrences = new HashMap<>();
        nodes = new ArrayList<>();

        //add initial dummy node so other elements are indexed properly
        nodes.add(new Node(1, "/"));
        //add initial dummy element so other elements are indexed properly
        automaton.addElement(new Element("dummy"));

        System.out.println("Building the Tree Path Automaton");
        System.out.println("walking the XML tree");
        walkTree(document.getRootElement());
        buildSJT(0);
        System.out.println("build NKA");
        buildNDelta();
        System.out.println("build DKA");
        determinize();
        System.out.println("Done.");
        return automaton;
    }

    public void serialize(String path) throws FileNotFoundException, IOException {
        FileOutputStream fout = new FileOutputStream(path);
        ObjectOutputStream oos = new ObjectOutputStream(fout);
        oos.writeObject(this);
    }

    /**
     * Deserealize the Tree Path Automaton from given file.
     *
     * @param path path to file with serialized TreePathAutomaton object
     * @return resulting TreePathAutomaton object
     */
    public TreePathsAutomaton deserealize(String path) {
        System.out.println("Loading the automaton from: " + path);
        ObjectInputStream objectInputStream;
        try {
            objectInputStream = new ObjectInputStream(new FileInputStream(path));
            automaton = (TreePathsAutomaton) objectInputStream.readObject();
        } catch (IOException | ClassNotFoundException ex) {
            Logger.getLogger(AutomatonFactory.class.getName()).log(Level.SEVERE, null, ex);
        }
        return automaton;
    }

    //walk the XML Tree in preorder fashion, initialize working structures
    //and associate the Element objects with their preorder identifiers.
    private void walkTree(Element e) {
        if (e == null) {
            return;
        }
        //process actual element
        List<Element> children = e.getChildren();
        nodes.add(new Node(children.size(), e.getName()));
        updateOccurrences(e, nodes.size() - 1);
        automaton.addElement(e);
        //process its children (in document order)
        children.stream().forEach((child) -> {
            walkTree(child);
        });
    }

    private void updateOccurrences(Element e, int id) {
        TreeSet<Integer> occ = occurrences.get(e.getName());

        if (occ == null) {
            occ = new TreeSet<>();
            occurrences.put(e.getName(), occ);
        }
        occ.add(id);
    }

    //build subtree jump table
    private int buildSJT(int rootIndex) {
        int index = rootIndex + 1;
        for (int i = 1; i <= nodes.get(rootIndex).getArity(); i++) {
            index = buildSJT(index);
        }
        nodes.get(rootIndex).setSJT(index);
        return index;
    }

    //build nondeterministic transition function of TreePathsAutomaton
    private void buildNDelta() {
        int i;
        for (int q = 0; q < nodes.size() - 1; q++) {

            i = q + 1;
            //transitions for / axis
            for (int j = 0; j < nodes.get(q).getArity(); j++) {
                updateNDelta(q, new EdgeLabel(nodes.get(i).getLabel(),
                        EdgeLabel.Axis.CHILD), i);
                i = nodes.get(i).getSJT();
            }
            //transitions for // axis
            for (String label : occurrences.keySet()) {
                for (Integer id : occurrences.get(label)) {
                    if (id > q && id < nodes.get(q).getSJT()) {
                        updateNDelta(q, new EdgeLabel(nodes.get(id).getLabel(),
                                EdgeLabel.Axis.DESCENDANT_OR_SELF), id);
                    }
                }
            }
        }
    }

    private void updateNDelta(int q1, EdgeLabel label, int q2) {
        HashMap<EdgeLabel, TreeSet<Integer>> trans = nDelta.get(q1);
        TreeSet<Integer> goals;

        // no transition for q1 yet
        if (trans == null) {
            trans = new HashMap<>();
            goals = new TreeSet<>();
            trans.put(label, goals);
            nDelta.put(q1, trans);
        } // some transitions for q1 already defined
        else {
            goals = trans.get(label);
            // no transition for label yet
            if (goals == null) {
                goals = new TreeSet<>();
                trans.put(label, goals);
            }
        }
        goals.add(q2);
    }

    //determinize the TreePathsAutomaton
    public void determinize() {
        //queue of unprocessed states
        LinkedList<State> queue = new LinkedList<>();
        //new states being created 
        HashMap<EdgeLabel, State> newStates;
        State newState, state;

        newState = new State(0);
        queue.add(newState);
        automaton.addState(newState);

        while (!queue.isEmpty()) {
            state = queue.remove();
            newStates = processState(state);
            processNewStates(newStates, queue, state);
            newStates.clear();
        }
    }

    private HashMap<EdgeLabel, State> processState(State state) {
        HashMap<EdgeLabel, State> newStates = new HashMap<>();
        HashMap<EdgeLabel, TreeSet<Integer>> trans;
        State newState;

        for (Integer id : state.getDSubset()) {
            if (nodes.get(id).getArity() != 0) {
                trans = nDelta.get(id);
                for (EdgeLabel label : trans.keySet()) {
                    newState = newStates.get(label);
                    if (newState == null) {
                        newState = new State();
                        newStates.put(label, newState);
                    }
                    newState.updateDSubset(trans.get(label));
                }
            }
        }
        return newStates;
    }

    private void processNewStates(HashMap<EdgeLabel, State> newStates,
            LinkedList<State> queue, State state) {
        State newState;
        for (EdgeLabel label : newStates.keySet()) {
            newState = newStates.get(label);
            State stateRef = automaton.getState(newState);
            if (stateRef == null) {
                queue.add(newState);
                automaton.addState(newState);
                stateRef = newState;
            }
            automaton.updateDelta(state, label, stateRef);
        }
    }

    /**
     * Print the working structures and nondetrministic TreePathsAutomaton to
     * stdout.
     */
    public void dump() {
        System.out.println("Working structures:");
        nodes.stream().forEach((node) -> {
            System.out.println(node.getLabel() + ":" + node.getArity()
                    + ":" + node.getSJT());
        });

        System.out.println("Occurences");
        occurrences.keySet().forEach((label) -> {
            System.out.print(label + ": {");
            occurrences.get(label).forEach((id) -> {
                System.out.print(id + ",");
            });
            System.out.print("");
            System.out.println("}");
        });

        System.out.println("\nNFA:\n");

        nDelta.keySet().forEach((q) -> {
            HashMap<EdgeLabel, TreeSet<Integer>> trans = nDelta.get(q);
            trans.keySet().forEach((label) -> {
                System.out.print("delta(" + q + "," + label + ")={");
                trans.get(label).forEach((id) -> {
                    System.out.print(id + ",");
                });
                System.out.println("}");
            });
        });
    }

    private class Node implements Serializable {

        int arity;
        String label;
        int sjt; //id of the first following-sibling

        protected Node(int arity, String label) {
            this.arity = arity;
            this.label = label;
        }

        protected void setSJT(int follower) {
            this.sjt = follower;
        }

        protected int getArity() {
            return arity;
        }

        protected String getLabel() {
            return label;
        }

        protected int getSJT() {
            return sjt;
        }
    };
}
