package cz.cvut.fit.xml.index.tpa;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.TreeSet;
import org.jdom2.Element;

/**
 * A TreePathsAutomaton is designed to process a significant fragment of XPath
 * queries, which may use any combination of child (/) and descendant-or-self
 * (//) axis.
 *
 * @author sestaeli
 */
public class TreePathsAutomaton implements Serializable {

    //list of XML results accessible by their preorder id
    private final ArrayList<Element> elements;
    //deterministic transition function
    private final HashMap<State, HashMap<EdgeLabel, State>> dDelta;
    //set of states to prevent duplications in the transition function
    private final LinkedList<State> states;

    /**
     * Initializes new <code>TreePathsAutomaton</code>.
     */
    TreePathsAutomaton() {
        elements = new ArrayList<>();
        states = new LinkedList<>();
        dDelta = new HashMap<>();
    }

    /**
     *
     * Serialize the TreePathAutomaton object into file by given path.
     *
     * @param path path to a file where TreePathAutomaton object will be
     * serialized
     * @throws FileNotFoundException if the file exists but is a directory
     * rather than a regular file, does not exist but cannot be created, or
     * cannot be opened for any other reason
     * @throws IOException Any exception thrown by the underlying OutputStream.
     */
    public void serialize(String path) throws FileNotFoundException, IOException {
        FileOutputStream fout = new FileOutputStream(path);
        ObjectOutputStream oos = new ObjectOutputStream(fout);
        oos.writeObject(this);
    }

    /**
     * Associate the element with its insertion order identifier.
     *
     * @param e XML element
     */
    void addElement(Element e) {
        elements.add(e);
    }

    /**
     * Save given state.
     *
     * @param s State
     */
    void addState(State s) {
        states.add(s);
    }

    /**
     * Verify whether given state already exists in the automaton.
     *
     * @param s State to be verified
     * @return state equals to s if s already defined or null otherwise.
     */
    State getState(State s) {
        int index = states.indexOf(s);
        return (index == -1) ? null : states.get(index);
    }

    /**
     * Update deterministic transition function by adding \u03B4(q1,label)=q2.
     *
     * @param q1 start transition state
     * @param label edge label
     * @param q2 end transition state
     */
    void updateDelta(State q1, EdgeLabel label, State q2) {
        HashMap<EdgeLabel, State> trans = dDelta.get(q1);

        // no transition for q1 yet
        if (trans == null) {
            trans = new HashMap<>();
            dDelta.put(q1, trans);
        }
        trans.put(label, q2);
    }

    private LinkedList<EdgeLabel> parse(String query) {
        LinkedList<EdgeLabel> inputs = new LinkedList<>();
        String[] tokens = query.split("//");
        for (String token : tokens) {
            if (!token.isEmpty()) {
                if (!token.contains("/")) {
                    inputs.add(new EdgeLabel(token,
                            EdgeLabel.Axis.DESCENDANT_OR_SELF));
                } else {
                    String[] t = token.split("/");
                    if (!t[0].isEmpty()) {
                        inputs.add(new EdgeLabel(t[0],
                                EdgeLabel.Axis.DESCENDANT_OR_SELF));
                    }
                    for (int i = 1; i < t.length; i++) {
                        inputs.add(new EdgeLabel(t[i], EdgeLabel.Axis.CHILD));
                    }
                }
            }
        }
        return inputs;
    }

    /**
     * Evaluate an XPath query.
     *
     * @param query XPath query
     * @return list of occurrences (d-subset) of XML elements satisfying the
     * query
     */
    public TreeSet<Integer> getDSubset(String query) {
        LinkedList<EdgeLabel> inputs = parse(query);
        //start at initial state
        State state = states.getFirst();
        while (!inputs.isEmpty() && state != null) {
            EdgeLabel label = inputs.remove();
            state = dDelta.get(state).get(label);
        }
        return (state == null) ? null : state.getDSubset();
    }

    /**
     * Maps d-subset (preorder ids) to appropriate elements.
     *
     * @param dsubset d-subset
     * @return list of XML elements corresponding to given d-subset.
     */
    public LinkedList<Element> evaluate(TreeSet<Integer> dsubset) {
        LinkedList<Element> results = new LinkedList<>();
        dsubset.stream().forEach((id) -> {
            results.add(elements.get(id));
        });
        return results;
    }

    /**
     * Evaluate an XPath query.
     *
     * @param query XPath query
     * @return list of XML elements satisfying the query
     */
    public LinkedList<Element> evaluate(String query) {
        LinkedList<Element> results = new LinkedList<>();
        LinkedList<EdgeLabel> inputs = parse(query);
        //start at initial state
        State state = states.getFirst();
        while (!inputs.isEmpty() && state != null) {
            EdgeLabel label = inputs.remove();
            state = dDelta.get(state).get(label);
        }
        if (state != null) {
            for (Integer id : state.getDSubset()) {
                results.add(elements.get(id));
            }
        }
        return results;
    }

    /**
     * Print the TreePathsAutomaton to stdout.
     */
    public void dump() {
        System.out.println("TREE PATH AUTOMATON M=(Q,T,\u03B4,[0],F)");
        //states
        System.out.print("Q = {");
        //states.stream().forEach((s) -> {
        //    System.out.print(s + ",");
        //});
        System.out.println("}");
        System.out.println("|Q| = " + states.size());
        //alphabet
        System.out.println("T = (/ + //).(e1 + e2 + ... e_k)");
        //final states
        System.out.println("F = Q - {[0]}");
        //delta

        System.out.println("\u03B4:");
        int transitions = 0;
        for (State q : dDelta.keySet()) {
            HashMap<EdgeLabel, State> trans = dDelta.get(q);
            transitions += trans.keySet().size();
        }
        System.out.println("transitions: " + transitions);

    }
}
